/** 
 * XMLSec library
 *
 * Simple Keys Manager
 * 
 * See Copyright for the status of this software.
 * 
 * Author: Aleksey Sanin <aleksey@aleksey.com>
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include <xmlsec/xmlsec.h>
#include <xmlsec/xmltree.h>
#include <xmlsec/keys.h>
#include <xmlsec/keysInternal.h>
#include <xmlsec/transforms.h>
#include <xmlsec/transformsInternal.h>
#include <xmlsec/keysmngr.h>


/**
 * Simple Keys Manager
 */

struct _xmlSecKeyMngr {
    xmlSecKeyPtr	*keys;
    size_t		current;
    size_t		max;
};

#define XMLSEC_SIMPLEKEYMNGR_DEFAULT			16

					 
/**
 * xmlSecSimpleKeyMngrCreate
 *
 *
 *
 */
xmlSecSimpleKeyMngrPtr	
xmlSecSimpleKeyMngrCreate(void) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrCreate";
    xmlSecSimpleKeyMngrPtr mngr;
        
    mngr = (xmlSecSimpleKeyMngrPtr)xmlMalloc(sizeof(struct _xmlSecKeyMngr));
    if(mngr == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to allocate struct _xmlSecKeyMngr\n",
	    func);	
#endif 	    
	return(NULL);
    }
    memset(mngr, 0, sizeof(struct _xmlSecKeyMngr)); 
            
    return(mngr);
}

/**
 * xmlSecSimpleKeyMngrDestroy
 *
 *
 *
 */
void
xmlSecSimpleKeyMngrDestroy(xmlSecSimpleKeyMngrPtr mngr) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrDestroy";

    if(mngr == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr is null\n",
	    func);	
#endif 	    
	return;    
    }
    
    if(mngr->keys != NULL) {
	size_t i;
	
	for(i = 0; i < mngr->current; ++i) {
	    if(mngr->keys[i] != NULL) {
		xmlSecKeyDestroy(mngr->keys[i], 1); /* force key deletion! */
	    }
	}
	memset(mngr->keys, 0, mngr->max * sizeof(xmlSecKeyPtr));
	xmlFree(mngr->keys);
    }
    memset(mngr, 0, sizeof(struct _xmlSecKeyMngr)); 
    xmlFree(mngr);
}

/**
 * xmlSecSimpleKeyMngrLoad
 * @mngr:
 * @uri:
 *
 */
int
xmlSecSimpleKeyMngrLoad(xmlSecSimpleKeyMngrPtr mngr, const char *uri, int strict) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrLoad";
    xmlSecKeysReadContext keysContext;
    xmlDocPtr doc;
    xmlNodePtr root;
    xmlNodePtr cur;
    xmlSecKeyPtr key;
    int ret;
    
    if((mngr == NULL) || (uri == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr or uri is null\n",
	    func);	
#endif 	    
	return(-1); 
    }
    
    doc = xmlParseFile(uri);
    if(doc == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to load keys from \"%s\"\n", 
	    func, uri);	
#endif
	return(-1);
    }
    
    root = xmlDocGetRootElement(doc);
    if(!xmlSecCheckNodeName(root, BAD_CAST "Keys", xmlSecDSigNs)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: bad root node\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(-1);
    }
    
    memset(&keysContext, 0, sizeof(keysContext));
    keysContext.allowedOrigins = xmlSecKeyOriginKeyValue;
    keysContext.maxRetrievals = xmlSecKeyInifiteRetrivals;
    keysContext.findKeyCallback = xmlSecSimpleKeyMngrFindKey;
    keysContext.findKeyContext = mngr;
    
    cur = xmlSecGetNextElementNode(root->children);
    while(xmlSecCheckNodeName(cur, BAD_CAST "KeyInfo", xmlSecDSigNs)) {  
	key = xmlSecKeyInfoRead(cur, xmlSecTransformUnknown, 
				xmlSecKeyTypeAny, &keysContext);
	if(key == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to read KeyInfo\n", 
		func);	
#endif
	    if(strict) {
		xmlFreeDoc(doc);
		return(-1);	
	    }
	} else {
	    ret = xmlSecSimpleKeyMngrAddKey(mngr, key);
	    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    		xmlGenericError(xmlGenericErrorContext,
		    "%s: failed to add key\n", 
		    func);	
#endif
		xmlSecKeyDestroy(key, 0);
		xmlFreeDoc(doc);
		return(-1);	
	    }
	}
        cur = xmlSecGetNextElementNode(cur->next);
    }
    
    if(cur != NULL) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: unexpected node found\n", 
	    func);	
#endif
	xmlFreeDoc(doc);
	return(-1);	    
    }
    
    xmlFreeDoc(doc);
    return(0);
}

/**
 * xmlSecSimpleKeyMngrSave
 * @mngr:
 * @filename:
 * @type:
 */
int
xmlSecSimpleKeyMngrSave(const xmlSecSimpleKeyMngrPtr mngr, 
			const char *filename, xmlSecKeyType type) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrSave";
    xmlSecKeysReadContext keysContext;
    xmlDocPtr doc;
    xmlNodePtr root;
    xmlNodePtr cur;
    int ret;
    size_t i;
    
    if((mngr == NULL) || (filename == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr or filename is null\n",
	    func);	
#endif 	    
	return(-1);    
    }

    /* create doc */
    doc = xmlNewDoc(BAD_CAST "1.0");
    if(doc == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create new doc\n",
	    func);	
#endif
	return(-1);
    }
    
    /* create root node "Keys" */
    root = xmlNewDocNode(doc, NULL, BAD_CAST "Keys", NULL); 
    if(root == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to create root doc node\n",
	    func);	
#endif
	xmlFreeDoc(doc);
	return(-1);
    }
    xmlDocSetRootElement(doc, root);
    if(xmlNewNs(root, xmlSecDSigNs, NULL) == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to add ns to node\n",
	    func);	
#endif
	xmlFreeDoc(doc); 
	return(-1);
    }
    memset(&keysContext, 0, sizeof(keysContext));
    keysContext.allowedOrigins = xmlSecKeyOriginKeyValue;
    keysContext.maxRetrievals = xmlSecKeyInifiteRetrivals;
    keysContext.findKeyCallback = xmlSecSimpleKeyMngrFindKey;
    keysContext.findKeyContext = mngr;
    
    for(i = 0; i < mngr->current; ++i) {
	cur = xmlSecAddChild(root, BAD_CAST "KeyInfo", NULL);
	if(cur == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to ad KeyInfo node\n",
		func);	
#endif
	    xmlFreeDoc(doc); 
	    return(-1);
	}
	
	if(xmlSecAddChild(cur, BAD_CAST "KeyName", NULL) == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add KeyName node\n",
		func);	
#endif
	    xmlFreeDoc(doc); 
	    return(-1);
	}

	if(xmlSecAddChild(cur, BAD_CAST "KeyValue", NULL) == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to add KeyValue node\n",
		func);	
#endif
	    xmlFreeDoc(doc); 
	    return(-1);
	}

	ret = xmlSecKeyInfoWrite(cur, mngr->keys[i], type, &keysContext);
	if(ret < 0) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to write KeyInfo node\n",
		func);	
#endif
	    xmlFreeDoc(doc); 
	    return(-1);
	}		
    }    

    /* now write result */
    ret = xmlSaveFormatFile(filename, doc, 1);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
    	xmlGenericError(xmlGenericErrorContext,
	    "%s: failed to write file \"%s\"\n", 
	    func, filename);	
#endif
	xmlFreeDoc(doc); 
	return(-1);
    }	   
    
    xmlFreeDoc(doc);
    return(0);
}

/**
 * xmlSecSimpleKeyMngrFindKey
 * @context:
 * @id:
 * @type:
 * @name:
 *
 * Lookups the first key that does match the given criteria
 */
xmlSecKeyPtr
xmlSecSimpleKeyMngrFindKey(void *context, xmlSecKeyId id,
			xmlSecKeyType type, const xmlChar *name) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrFindKey";
    xmlSecSimpleKeyMngrPtr mngr;
    xmlSecKeyPtr key;
    size_t i;
    
    
    if(context == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: context is null\n",
	    func);	
#endif 	    
	return(NULL);    
    }
    mngr = (xmlSecSimpleKeyMngrPtr)context;
    
    for(i = 0; i < mngr->current; ++i) {
	key = mngr->keys[i];
	
	if(key == NULL) continue;
	if((id != xmlSecKeyIdUnknown) && (id != key->id)) continue;
	if((type != xmlSecKeyTypeAny) && (key->type != type) &&
	   key->type != xmlSecKeyTypePrivate) continue;
	if((name != NULL) && (!xmlStrEqual(key->name, name))) continue;

	return(key);
    }
        
    return(NULL);
}

/**
 * xmlSecSimpleKeyMngrAddKey:
 * @mngr:
 * @key:
 *
 * Adds new key to the key manager
 */
int	
xmlSecSimpleKeyMngrAddKey(xmlSecSimpleKeyMngrPtr mngr, xmlSecKeyPtr key) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrAddKey";

    if((mngr == NULL) || (key == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr or key is null\n",
	    func);	
#endif 	    
	return(-1);    
    }
    
    if(mngr->max == 0) {
	mngr->keys = (xmlSecKeyPtr *) xmlMalloc(XMLSEC_SIMPLEKEYMNGR_DEFAULT *
					    sizeof(xmlSecKeyPtr));
	if(mngr->keys == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to allocate %d keys pointers\n",
		func, XMLSEC_SIMPLEKEYMNGR_DEFAULT);	
#endif 	    
	    return(-1);
	}
	memset(mngr->keys, 0, XMLSEC_SIMPLEKEYMNGR_DEFAULT * sizeof(xmlSecKeyPtr)); 
	mngr->max = XMLSEC_SIMPLEKEYMNGR_DEFAULT;
    } else if(mngr->current == mngr->max) {
	xmlSecKeyPtr *newKeys;
	size_t newMax;
	
	newMax = mngr->max * 2;
	newKeys = (xmlSecKeyPtr *) xmlRealloc(mngr->keys, newMax * sizeof(xmlSecKeyPtr));
	if(newKeys == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to allocate %d keys pointers\n",
		func, newMax);	
#endif 	    
	    return(-1);	
	}
	mngr->max = newMax;
	mngr->keys = newKeys;
    }
    
    mngr->keys[(mngr->current)++] = key;
    key->origin |= xmlSecKeyOriginKeyManager;
    return(0);
}

/**
 * xmlSecSimpleKeyMngrLoadPrivateKey
 * @mngr:
 * @keyfile:
 * @keyPwd:
 * @keyPwdCallback:
 * @certfile:
 *
 */
xmlSecKeyPtr
xmlSecSimpleKeyMngrLoadPrivateKey(xmlSecSimpleKeyMngrPtr mngr, 
				const char *keyfile, const char *keyPwd,
				pem_password_cb *keyPwdCallback) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrLoadPrivateKey";
    xmlSecKeyPtr key = NULL;
    EVP_PKEY *pKey = NULL;    
    FILE *f;
    int ret;
    
    if((mngr == NULL) || (keyfile == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr or key file is null\n",
	    func);	
#endif 	    
	return(NULL);    
    }
    
    f = fopen(keyfile, "r");
    if(f == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to open file \"%s\"\n",
	    func, keyfile);	
#endif 	    
	return(NULL);    
    }
    
    pKey = PEM_read_PrivateKey(f, NULL, keyPwdCallback, (void*)keyPwd);
    if(pKey == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to read key file \"%s\"\n",
	    func, keyfile);	
#endif 	    
	fclose(f);
	return(NULL);    
    }
     
    key = xmlSecKeyReadEvpKey(pKey, xmlSecTransformUnknown, xmlSecKeyOriginKeyManager);
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to get key\n",
	    func);	
#endif 	    
	EVP_PKEY_free(pKey);
        fclose(f);	
	return(NULL);
    }        
    EVP_PKEY_free(pKey);
    fclose(f);
    
    ret = xmlSecSimpleKeyMngrAddKey(mngr, key);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to add key to the keymanager\n",
	    func);	
#endif 	    
	xmlSecKeyDestroy(key, 1);
	return(NULL);
    }
    
    return(key);
}

/**
 * xmlSecSimpleKeyMngrLoadPublicKey
 * @mngr:
 * @keyfile:
 * @keyPwd:
 * @keyPwdCallback:
 *
 */
xmlSecKeyPtr
xmlSecSimpleKeyMngrLoadPublicKey(xmlSecSimpleKeyMngrPtr mngr,
				const char *keyfile, const char *keyPwd,
				pem_password_cb *keyPwdCallback) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrLoadPublicKey";
    xmlSecKeyPtr key = NULL;
    EVP_PKEY *pKey = NULL;        
    FILE *f;
    int ret;
    
    if((mngr == NULL) || (keyfile == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: mngr or key file is null\n",
	    func);	
#endif 	    
	return(NULL);    
    }
        
    f = fopen(keyfile, "r");
    if(f == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to open file \"%s\"\n",
	    func, keyfile);	
#endif 	    
	return(NULL);    
    }
    
    pKey = PEM_read_PUBKEY(f, NULL, keyPwdCallback, (void*)keyPwd);
    if(pKey == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to read key file \"%s\"\n",
	    func, keyfile);	
#endif 	    
	fclose(f);
	return(NULL);    
    }
     
    key = xmlSecKeyReadEvpKey(pKey, xmlSecTransformUnknown, xmlSecKeyOriginKeyManager);
    if(key == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to get key\n",
	    func);	
#endif 	    
	EVP_PKEY_free(pKey);
        fclose(f);	
	return(NULL);
    }    
    EVP_PKEY_free(pKey);
    fclose(f);
    
    ret = xmlSecSimpleKeyMngrAddKey(mngr, key);
    if(ret < 0) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to add key to the keymanager\n",
	    func);	
#endif 	    
	xmlSecKeyDestroy(key, 1);
	return(NULL);
    }
    
    return(key);
}

/**
 * xmlSecSimpleKeyMngrAddCertToKey:
 * @key: 
 * @certfile:
 *
 * Loads x509 cert from file and adds it to the key certs
 */
int
xmlSecSimpleKeyMngrAddCertToKey(xmlSecKeyPtr key, const char *certfile) {
    static const char func[] _UNUSED_VARIABLE_ = "xmlSecSimpleKeyMngrAddCertToKey";

#ifndef XMLSEC_NO_X509    
    X509 *cert;
    FILE *f;

    if((key == NULL) || (certfile == NULL)) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: key or cert file is null\n",
	    func);	
#endif 	    
	return(-1);    
    }

    if(key->x509Chain == NULL) {
        /* init chain */
	key->x509Chain = sk_X509_new_null();
        if(key->x509Chain == NULL) {
#ifdef XMLSEC_DEBUG
    	    xmlGenericError(xmlGenericErrorContext,
		"%s: failed to create x509 certs stack\n",
		func);	
#endif
	    return(-1);
	}
    }

    f = fopen(certfile, "r");
    if(f == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to open file \"%s\"\n",
	    func, certfile);	
#endif 	    
	return(-1);    
    }
    
    cert = PEM_read_X509_AUX(f, NULL, NULL, NULL);
    if(cert == NULL) {
#ifdef XMLSEC_DEBUG
        xmlGenericError(xmlGenericErrorContext,
	    "%s: unable to read cert file \"%s\"\n",
	    func, certfile);	
#endif 	    
	fclose(f);
	return(-1);    
    }
    	
    sk_X509_push(key->x509Chain, cert);
    fclose(f);
    return(0);
#else /* XMLSEC_NO_X509 */
#ifdef XMLSEC_DEBUG
    xmlGenericError(xmlGenericErrorContext,
	"%s: x509 support disabled\n",
	func);	
#endif 	    
    return(-1);
#endif /* XMLSEC_NO_X509 */
}
